                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*version 5: functional continuations*/
                     /*-----------------------------------*/
                     /*               grammar             */
                     /*-----------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SlipGrammar.h"

/*------------------------------ private macros ------------------------------*/

#define chunk_size(TYP)                                                        \
  (sizeof(TYP) / sizeof(CEL_type) - 1)

#define make_chunk_with_offset(TYP, SIZ)                                       \
  (TYP##_type)Memory_Make_Chunk(TYP##_tag, TYP##_size + SIZ)

#define make_chunk(TYP)                                                        \
  make_chunk_with_offset(TYP, 0)

/*------------------------------ private functions ---------------------------*/

static BYT_type member_of(EXP_type Expression,
                          TAG_type Tag)
  { BYT_type tag;
    tag = Tag_of(Expression);
    return (tag == Tag); }

static UNS_type string_to_expression_size(TXT_type Raw_string)
  { return strlen(Raw_string) / sizeof(CEL_type) + 1; }

/*-------------------------------------Tag -----------------------------------*/

TAG_type Tag_of(EXP_type Expression)
  { return (TAG_type)Memory_Get_Tag((PTR_type)Expression); }

/*------------------------------ CFN declarations ----------------------------*/

static const UNS_type CFN_size = chunk_size(CFN);

BYT_type is_CFN(EXP_type Expression)
  { return member_of(Expression,
                     CFN_tag); }

CFN_type make_CFN(CCC_type C_function)
  { CFN_type c_function;
    c_function = make_chunk(CFN);
    c_function->ccc = C_function;
    return c_function; }

/*------------------------------ CHA declarations ----------------------------*/

static const UNS_type CHA_size = chunk_size(CHA);

BYT_type is_CHA(EXP_type Expression)
  { return member_of(Expression,
                     CHA_tag); }

CHA_type make_CHA(CHR_type Raw_character)
  { CHA_type character;
    character = make_chunk(CHA);
    character->chr = Raw_character;
    return character; }

/*------------------------------ CNT declarations ----------------------------*/

static const UNS_type CNT_size = chunk_size(CNT);

BYT_type is_CNT(EXP_type Expression)
  { return member_of(Expression,
                     CNT_tag); }

CNT_type make_CNT(CFN_type C_function,
                  CNT_type Continuation,
                  UNS_type Size)
  { CNT_type continuation;
    continuation = make_chunk_with_offset(CNT,
                                          Size);
    continuation->cfn = C_function;
    continuation->cnt = Continuation;
    return continuation; }

/*------------------------------ ENV declarations ----------------------------*/

static const UNS_type ENV_size = chunk_size(ENV);

BYT_type is_ENV(EXP_type Expression)
  { return member_of(Expression,
                     ENV_tag); }

ENV_type make_ENV(SYM_type Variable,
                  EXP_type Value,
                  ENV_type Environment)
  { ENV_type environment;
    environment = make_chunk(ENV);
    environment->var = Variable;
    environment->val = Value;
    environment->env = Environment;
    return environment; }

/*------------------------------ FLS declarations ----------------------------*/

static const UNS_type FLS_size = chunk_size(FLS);

BYT_type is_FLS(EXP_type Expression)
  { return member_of(Expression,
                     FLS_tag); }

FLS_type make_FLS(NIL_type)
  { FLS_type false;
    false = make_chunk(FLS);
    return false; }

/*------------------------------ NAT declarations ----------------------------*/

static const UNS_type NAT_size = chunk_size(NAT);

BYT_type is_NAT(EXP_type Expression)
  { return member_of(Expression,
                     NAT_tag); }

NAT_type make_NAT(FUN_type Raw_function,
                  TXT_type Name)
  { NAT_type native;
    UNS_type size;
    size = string_to_expression_size(Name);
    native = make_chunk_with_offset(NAT,
                                    size);
    native->fun = Raw_function;
    strcpy(native->nam,
           Name);
    return native; }

/*------------------------------ NBR declarations ----------------------------*/

static const UNS_type NBR_size = chunk_size(NBR);

BYT_type is_NBR(EXP_type Expression)
  { return member_of(Expression,
                     NBR_tag); }

NBR_type make_NBR(LNG_type Raw_number)
  { NBR_type number;
    number = make_chunk(NBR);
    number->lng = Raw_number;
    return number; }

/*------------------------------ NUL declarations ----------------------------*/

static const UNS_type NUL_size = chunk_size(NUL);

BYT_type is_NUL(EXP_type Expression)
   { return member_of(Expression,
                     NUL_tag); }

NUL_type make_NUL(NIL_type)
  { NUL_type null;
    null = make_chunk(NUL);
    return null; }

/*------------------------------ PAI declarations ----------------------------*/

static const UNS_type PAI_size = chunk_size(PAI);

BYT_type is_PAI(EXP_type Expression)
  { return member_of(Expression,
                     PAI_tag); }

PAI_type make_PAI(EXP_type Car,
                  EXP_type Cdr)
  { PAI_type pair;
    pair = make_chunk(PAI);
    pair->car = Car;
    pair->cdr = Cdr;
    return pair; }

/*------------------------------ PRC declarations ----------------------------*/

static const UNS_type PRC_size = chunk_size(PRC);

BYT_type is_PRC(EXP_type Expression)
  { return member_of(Expression,
                     PRC_tag); }

PRC_type make_PRC(SYM_type Name,
                  EXP_type Parameters,
                  EXP_type Body,
                  ENV_type Closure)
  { PRC_type procedure;
    procedure = make_chunk(PRC);
    procedure->nam = Name;
    procedure->par = Parameters;
    procedure->bod = Body;
    procedure->clo = Closure;
    return procedure; }

/*------------------------------ REA declarations ----------------------------*/

static const UNS_type REA_size = chunk_size(REA);

BYT_type is_REA(EXP_type Expression)
  { return member_of(Expression,
                     REA_tag); }

REA_type make_REA(FLO_type Raw_real)
  { REA_type real;
    real = make_chunk(REA);
    real->flo = Raw_real;
    return real; }

/*------------------------------ STR declarations ----------------------------*/

static const UNS_type STR_size = chunk_size(STR);

BYT_type is_STR(EXP_type Expression)
  { return member_of(Expression,
                     STR_tag); }

STR_type make_STR(TXT_type Raw_string)
  { STR_type string;
    UNS_type size;
    size = string_to_expression_size(Raw_string);
    string = make_chunk_with_offset(STR,
                                    size);
    strcpy(string->txt,
           Raw_string);
    return string; }

/*------------------------------ SYM declarations ----------------------------*/

static const UNS_type SYM_size = chunk_size(SYM);

BYT_type is_SYM(EXP_type Expression)
  { return member_of(Expression,
                     SYM_tag); }

SYM_type make_SYM(TXT_type Raw_string)
  { SYM_type symbol;
    UNS_type size;
    size = string_to_expression_size(Raw_string);
    symbol = make_chunk_with_offset(SYM,
                                    size);
    strcpy(symbol->txt,
           Raw_string);
    return symbol; }

/*------------------------------ TRU declarations ----------------------------*/

static const UNS_type TRU_size = chunk_size(TRU);

BYT_type is_TRU(EXP_type Expression)
  { return member_of(Expression,
                     TRU_tag); }

TRU_type make_TRU(NIL_type)
  { TRU_type true;
    true = make_chunk(TRU);
    return true; }

/*------------------------------ USP declarations ----------------------------*/

static const UNS_type USP_size = chunk_size(USP);

BYT_type is_USP(EXP_type Expression)
  { return member_of(Expression,
                     USP_tag); }

USP_type make_USP(NIL_type)
  { USP_type unspecified;
    unspecified = make_chunk(USP);
    return unspecified; }

/*------------------------------ VEC declarations ----------------------------*/

static const UNS_type VEC_size = 0;

BYT_type is_VEC(EXP_type Expression)
  { return member_of(Expression,
                     VEC_tag); }

VEC_type make_VEC(UNS_type Size)
  { VEC_type vector;
    vector = (VEC_type)make_chunk_with_offset(VEC,
                                              Size);
    return vector; }

UNS_type size_VEC(VEC_type Vector)
  { return (UNS_type) Memory_Get_Size((PTR_type)Vector); }